/* 
 * Copyright (c) 2012, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/*
 *  ======== Global.c ========
 */

/* This define must precede inclusion of any xdc header files */
#define Registry_CURDESC ti_sdo_ce_osal_Global_desc

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Registry.h>

#include <ti/sdo/ce/global/CESettings.h>

#include <stdlib.h>

#include <ti/sdo/ce/osal/Global.h>

#include <unistd.h> /* For getpid() */

/* list of functions to be called at exit */

typedef struct ExitFxnElem {
    Fxn                 fxn;
    struct ExitFxnElem *next;
} ExitFxnElem;

/* REMINDER: if you add an initialized static var, reinitialize it at cleanup */

static Bool         curInit          = FALSE;

static ExitFxnElem *exitFxnList      = NULL;
static Bool         doRegisterAtExit = TRUE;


extern String Global_buildInfo[];  /* generated by Global.xdt */

Registry_Desc ti_sdo_ce_osal_Global_desc;

static Int regInit = 0;     /* Registry_addModule() called */

/*
 *  ======== Global_abort ========
 *  TODO: (jeh) This can be removed
 */
Void Global_abort(String fmt, ...)
{
    va_list vargs;

    System_printf("\n======== Global_abort =============\n");

    va_start(vargs, fmt);
    System_vprintf(fmt, vargs);
    va_end(vargs);

    System_exit(1);
}

/*
 *  ======== Global_atexit ========
 *  TODO: (jeh) Use System_atexit(System_AtexitHandler fxn), where
 *  typedef Void (*System_AtexitHandler)(Int);
 */
Bool Global_atexit(Fxn fxn)
{
    ExitFxnElem *exitFxnElem;

    Log_print1(Diags_ENTRY, "[+E] Global_atexit> Enter (fxn=0x%x)",
            (IArg)(fxn ));

    exitFxnElem = (ExitFxnElem *)Memory_alloc(NULL, sizeof(ExitFxnElem), 0,
            NULL);

    if (exitFxnElem == NULL) {
        return (FALSE);
    }

    exitFxnElem->fxn  = fxn;
    exitFxnElem->next = exitFxnList;
    exitFxnList       = exitFxnElem;

    /* register with the real system exit only once in the lifetime */
    if (doRegisterAtExit == TRUE) {
        System_atexit((System_AtexitHandler)Global_exit );
        doRegisterAtExit = FALSE;
    }

    return (TRUE);
}

/*
 *  ======== Global_exit ========
 *  calls cleanup functions for all the modules that
 *  scheduled it, then cleans up itself
 */
Void Global_exit(Void)
{
    ExitFxnElem *old;

    if (curInit == FALSE) {
        return;
    }

    Log_print0(Diags_ENTRY, "[+E] Global_exit> Enter");

    while( exitFxnList != NULL ) {
        Log_print1(Diags_USER2, "[+2] Global_exit> "
                "calling function *0x%x()...", (IArg)(exitFxnList->fxn ));
        exitFxnList->fxn();
        old = exitFxnList;
        exitFxnList = exitFxnList->next;
        Memory_free(NULL, old, sizeof(ExitFxnElem));
    }

    /* if the tracing module has its exit function, it must have been
     * the last one, so don't use trace from this point on
     */

    /* do the cleanup of this module */
    curInit          = FALSE;
    exitFxnList      = NULL;  /* unnecessary, NULL already */

    /* we do NOT reset doRegisterAtExit, as we don't want endless pairs of
     * CERuntime_init/exit to keep adding this func to the process' atexit
     * table. Autoexit is a curtesy feature only for those who call
     * CERuntime_init once.
     */
    /* doRegisterAtExit = TRUE; */
}

/*
 *  ======== Global_getCpuLoad ========
 */
Int Global_getCpuLoad(Void)
{
    return (-1);
}

/*
 *  ======== Global_getenv ========
 */
String Global_getenv(String name)
{
    return (getenv(name));
}

/*
 *  ======== Global_getProcessId ========
 */
UInt32 Global_getProcessId(Void)
{
    return ((UInt32)getpid());
}

/*
 *  ======== Global_init ========
 */
Void Global_init(Void)
{
    Int i;
    Registry_Result   result;

    /*
     *  No need to reference count for Registry_addModule(), since there
     *  is no way to remove the module.
     */
    if (regInit == 0) {
        /* Register this module for logging */
        result = Registry_addModule(&ti_sdo_ce_osal_Global_desc,
                Global_MODNAME);
        Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);

        if (result == Registry_SUCCESS) {
            /* Set the diags mask to the CE default */
            CESettings_init();
            CESettings_setDiags(Global_MODNAME);
        }
        regInit = 1;
    }

    Log_print0(Diags_ENTRY, "[+E] Global_init> Enter");

    if (curInit != TRUE) {
        curInit = TRUE;

        Log_print0(Diags_USER4, "[+4] Global_init> This program was built "
                "with the following packages:");
        for (i = 0; Global_buildInfo[i] != NULL; i++) {
            Log_print1(Diags_USER4, "[+4] %s", (IArg)(Global_buildInfo[i]));
        }
    }
}

/*
 *  ======== Global_setSpecialTrace ========
 */
Void Global_setSpecialTrace( String mask )
{
}
/*
 *  @(#) ti.sdo.ce.osal.linux; 2, 0, 1,1; 1-17-2012 23:15:34; /db/atree/library/trees/ce/ce-t03/src/ xlibrary

 */

